/*
Copyright 2003/2004, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.
Contributors: Marco Hutta

For documentation and questions please use the web site
http://www.scatterweb.net and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2003/2004
*/

#ifndef _H_MSP430ClockMESF_h
#define _H_MSP430ClockMESF_h

#define SDA_HIGH  (P5OUT |= 0x01)
#define SDA_LOW   (P5OUT &= 0xFE)
#define SCL_HIGH  (P5OUT |= 0x02)
#define SCL_LOW   (P5OUT &= 0xFD)
#define BUS_READ          0x9F
#define BUS_WRITE         0x9E
#define ACC_CSR           0xAC    
#define ACC_CLOCK         0xC0

void cl_start(void){
  P5DIR |= 0x03;
  SCL_LOW;  _NOP(); _NOP();
  SDA_HIGH; _NOP(); _NOP();
  SCL_HIGH; _NOP(); _NOP();
  SDA_LOW;  _NOP(); _NOP();
  SCL_LOW;  _NOP(); _NOP();
}

int cl_writeOnBus(char byte) {
  int i, ack;
  for(i=0;i<8;i++) {
    if(byte & 0x80) SDA_HIGH; else SDA_LOW;
    SCL_HIGH;
    byte = byte << 1;
    SCL_LOW;
  }
  P5DIR &= 0xFE;                             
  SCL_HIGH;
  if(P5IN & 0x01) ack = 0; else ack = 1;     
  SCL_LOW;
  P5DIR |= 0x01;                             
  return ack;
}

char cl_readFromBus(int ack) {
  int i;
  char byte = 0;
  P5DIR &= 0xFE;                            
  for(i=0;i<8;i++) {
    byte = byte << 1;                
    SCL_HIGH;
    if(P5IN & 0x01) byte |= 0x01; else byte &= 0xFE;
    SCL_LOW;
  }
  P5DIR |= 0x01;                            
  if(ack) SDA_LOW; else SDA_HIGH;
  SCL_HIGH;
  SCL_LOW;
  return byte;
}

void cl_stop(void){
  SCL_LOW;  _NOP(); _NOP();
  SDA_LOW;  _NOP(); _NOP();
  SCL_HIGH; _NOP(); _NOP();
  SDA_HIGH; _NOP(); _NOP();
  SCL_LOW;  _NOP(); _NOP();
  P5DIR &= ~0x03;
}

unsigned int getReg16bit(unsigned char acc,unsigned int bitmask){
  int config = 0;
  do cl_start();
  while(!cl_writeOnBus(BUS_WRITE)); 
  cl_writeOnBus(acc);
  cl_start(); 
  cl_writeOnBus(BUS_READ);
  config = cl_readFromBus(1);
  config = config << 8;
  config += cl_readFromBus(0);
  cl_stop();
  config &= bitmask;
  _NOP();
  _NOP();
  return config;
}

void setCSReg(unsigned char setting){
  do cl_start();
  while(!cl_writeOnBus(BUS_WRITE));               
  cl_writeOnBus(ACC_CSR);                         
  cl_writeOnBus(setting);
  cl_stop();
  _NOP();
  _NOP();
  _NOP();
  _NOP();
}

void resetCHBit(void){
  do cl_start();                 
  while(!cl_writeOnBus(BUS_WRITE)); 
  cl_writeOnBus(ACC_CLOCK);       
  cl_writeOnBus(0x00);             

  cl_writeOnBus(0x00); 
  cl_stop();         
}

void setDcoEsb(void) {
	#define DELTA  300               
	#define UP		 1                     
	#define DOWN     2                    
	unsigned int Compare, Oldcapture = 0, i, direction = 0;
	BCSCTL1 = 0x86;
	BCSCTL2 = 0x00;
	DCOCTL = 0xFF;
	for (i = 0xFFFF; i > 0; i--);       
	CCTL2 = CCIS0 + CM0 + CAP;           
	TACTL = TASSEL1 + TACLR + MC1;       
	while (1){
	    CCTL2 &= ~(CCIFG + COV);         
		while (!(CCTL2 & CCIFG));        
		CCTL2 &= ~CCIFG;                 
		Oldcapture = CCR2;               
	    for(i=0; i<4; i++){
			while (!(CCTL2 & CCIFG));    
			CCTL2 &= ~CCIFG;             
			Compare = CCR2;
		}
		Compare = Compare - Oldcapture;  
		if (DELTA == Compare){
			break;       
		}
		else if (DELTA < Compare){           
			if(direction == UP) break;        
			if(!direction) direction = DOWN;  
			DCOCTL--;
			if (DCOCTL == 0xFF){
				BCSCTL1--;    
			}
		}    
		else {                      
			if(direction == DOWN) break;     
			if(!direction) direction = UP;   
			DCOCTL++;
			if (DCOCTL == 0x00){
				BCSCTL1++;   
			}
		}
	}  
	CCTL2 = 0;                             
	TACTL = 0;                             
	#undef DELTA
	#undef UP
	#undef DOWN    
}

#endif
